/*
DTWLIB.CPP -- DTREND PROJECT
Author: Timo Ahonen
Version: 0.3
Revision date: 14.2.2002
Creation date: 3.3.2002


DTREND PROJECT
Project Manager: Juha Ylipekkala
Project group: Timo Ahonen, Tero Rontti, Juha Ylipekkala

----------------------------------------------------------------------------
REVISION HISTORY:

Version 0.1 
Initial version
Author: Timo Ahonen
Date: 14.2.2002
Known bugs/problems:
-Too few comments in the code
-Handling of certain invalid inputs:
	-Too short input vectors (length < 3)
	-Entering a invalid value of tIndex for DTCompareOne function
-TBD: what to do when params.wwidth < abs(observation.length-template.length)

Version 0.3
Authors; Timo Ahonen, Juha Ylipekkala
Date: 3.3.2002
- Changed the return type of DTWCompareOne to DTWOneResult* (A pointer to _single_ struct)
- Added some comments

*/

/*
#define DTWLIBDEBUG 1
*/

#include "stdafx.h"
#include "dtwlib.h"

#include <vector>
#include <algorithm>
#include <float.h>

//For debugging purposes
#ifdef DTWLIBDEBUG
#include <iostream>
#include <fstream>
#endif

using namespace std;

struct DTWTemplate {
  float * values;
  unsigned int length;
}; //struct

struct WarpingGridPoint {
	float costsofar;
	int lengthsofar;
	int previ;
	int prevj;
}; //struct

struct WPPoint {
	int x;
	int y;
}; //struct

WarpingGridPoint** createWarpingGrid(int x, int y, int n, int m, int o, float* observation, 
									 float* ttemplate, float (*dist)(int,int,float*,float*));

float sqrDist(int i, int j, float* observation, float* ttemplate);

float* calcDerivatives(float* sSeq, int slen);

bool allResultsElementCompare(DTWAllResultsElement res1, DTWAllResultsElement res2);

vector<DTWTemplate*> templates; //A global data structure containing pointers to templates

DTWOneResult* lastoneresult=NULL;
DTWAllResultsElement* lastallresults=NULL;
float* lasttemplate=NULL;




inline int maxint(int a, int b)
//Returns the maximum one of two integers
{
	return (a>b?a:b);
} //maxint

inline int maxint3(int a, int b, int c)
//Returns the maximum of three integers
{
	int temp=(a>b?a:b);
	return (temp>c?temp:c);
} //maxint3

inline int minint(int a, int b)
//Returns the minimum of two integers
{
	return(a<b?a:b);
} //minint


BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
    int i;
	
	#ifdef DTWLIBDEBUG	
	ofstream debug("dtwlibdebug.txt",ios_base::out|ios_base::app);
    #endif	
	
	switch (ul_reason_for_call)
	{

		case DLL_PROCESS_ATTACH:
			#ifdef DTWLIBDEBUG
			debug << "Process Attach";
			#endif
			break;
		case DLL_THREAD_ATTACH:
			break;
		case DLL_THREAD_DETACH:
			break;
		case DLL_PROCESS_DETACH: 
			#ifdef DTWLIBDEBUG
			debug << "Process Detach";
			#endif
			for(i=0;i<templates.size();i++) //Free the memory used by the templates
				DTDelTemplate(i);
			break;
    } //switch

	#ifdef DTWLIBDEBUG
	debug.close();
	#endif

    return TRUE;
} //DllMain

DTWLIB_API int DTAddTemplate(float* tSeq, int nSeq) 
//Adds a template into the internal data structures of the DLL.
//Parameters:
//float* tSeq - an array containing the template to be added
//int nSeq - length of tSeq
//Return value:
//int - an unique index number for the added template.
{
	DTWTemplate *newtemplate = new DTWTemplate;
	unsigned int i;
	unsigned int last_index;

	#ifdef DTWLIBDEBUG
	ofstream debug("componedebug.txt",ios_base::out|ios_base::app);
	debug << "Add template nSeq=" << nSeq << ": ";
	for(i=0;i<nSeq;i++)
		debug << tSeq[i] << " ";
	debug << endl;
	debug.close();
	#endif

	newtemplate->length=nSeq;
	newtemplate->values=new float[nSeq];
	for(i=0; i<nSeq; i++) //Copy the values of the template into own array
		newtemplate->values[i]=tSeq[i];
	templates.push_back(newtemplate); //Save a pointer to the new DTWTemplate struct.
  
	last_index=templates.size()-1; //Index number of the last (just added) template
	return last_index;
} //DTAddTemplate


DTWLIB_API int DTDelTemplate(int tIndex)
//Deletes the selected template
//Parameters:
//int tIndex - index number of the template to be deleted
//Return value:
//int - 0 if the deleting succeeded, 1 if template doesn't exist
{
	DTWTemplate *deltemplate;
	unsigned int retval=0;

	if((tIndex < templates.size()) && (templates[tIndex] != NULL))
	{
		deltemplate=templates[tIndex];
		delete [] deltemplate->values; //Delete the values array
		deltemplate->values=NULL; 
		delete deltemplate; //Delete the struct
		templates[tIndex] = NULL;
	} else
	{
		retval=1; //If template does not exist or points to NULL, return 1.
	}	

	return retval;
} //DTDelTemplate

DTWLIB_API DTWAllResultsElement* DTCompareAll(float* oSeq, int nSeq, 
						DTWParameters params, int* numresults)
//Compare the given observation against all templates in memory
//
//Parameters:
//float* oSeq - an array containing the observation
//int nSeq - length of oSeq
//DTWParameters params - a structure containing the parameters for the DTW algorithm
//int* numresult - the number of elements in the returned array will be stored into the memory location pointed by this pointer
//
//Return value:
//DTWAllResultsElement* - An array containing the results
{
	int i,j; //General variables for iteration
	vector<DTWAllResultsElement> results; //For storing the results temporarily
	DTWOneResult* thisresult; //The result of one comparison from DTCompareOne
	DTWAllResultsElement tempresult;  //Result of one comparison
	DTWAllResultsElement* returnarray;  //The array to be returned

	//Iterate through every possible template index number and if the template
	//in question exists, do the comparison using DTCompareOne function
	for(i=0;i<templates.size();i++)
		if (templates[i] != NULL)
		{
			thisresult=DTCompareOne(oSeq,nSeq,params,i);
			//Put the results into a STL vector of DTWAllResultsElement
			tempresult.tIndex=i;
			tempresult.dtwDistance=(*thisresult).dtwDistance;
			results.push_back(tempresult);
			//Free the memory used by the warping path array
			for(j=0;j<(*thisresult).pathLen;j++)
				delete[] (*thisresult).warpingPath[j];
			delete[] (*thisresult).warpingPath;
			delete thisresult;
			lastoneresult=NULL; //If this wasn't set to NULL, DTFreeResources would cause an error
		
		} //if
	
	//Sort the results vector so that template with least dtwDistance is first etc
	sort(results.begin(),results.end(),allResultsElementCompare);
	
	//Copy the contents of the vector into an array
	*numresults=results.size();

	returnarray=new DTWAllResultsElement[results.size()];
	for(i=0;i<results.size();i++)
		returnarray[i]=results[i];

	//Save a pointer for freeing the memory by DTFreeResources
	lastallresults=returnarray;
	
	return returnarray;
} //DTCompareAll


DTWLIB_API DTWOneResult* DTCompareOne(float* oSeq, int nSeq,
									 DTWParameters params, int tIndex)

//Compare the given observation against a chosen template
//
//Parameters:
//float* oSeq - an array containing the observation
//int nSeq - length of oSeq
//DTWParameters params - a structure containing the parameters for the DTW algorithm
//int tIndex - index number of the template
//
//Return value:
//DTWOneResult* - a pointer to structure containing the results of the comparison
{
	float *thisobs, *thistemp; //Pointers to the observation and template vectors used in comparison
	float thiscost,avgcost; //User for saving intermediate results when searching the best start and end points
	WarpingGridPoint** warpinggrid; //Pointer to the warping grid
	int x,y,a,b; //(x,y)=start point and (a,b)=end point of the warping path
	int sx,sy,sa,sb; //Start and end points of the best warping path so far
	int i,j; //Genearal variables for iteration etc
	int tlen; //length of the template vector;
	int **retpath; //The warping path array to be returned by the function
	int rplen; //Number of points on the warping path
	DTWOneResult* retresult; //The structure to be returned
	WPPoint temppoint; 
	vector<WPPoint> warpingpath; //The temporary vector used to construct the warping path
	
	
	#ifdef DTWLIBDEBUG
	ofstream debug("componedebug.txt",ios_base::out|ios_base::app);
	debug << "Compare one, tIndex=" << tIndex << " nSeq=" << nSeq << " wwidth=" << params.wwidth << " dmetrics=" << params.dmetrics << " seoffset=" << params.seoffset << endl;
	debug << "Observation: ";
	for(i=0;i<nSeq;i++)
		debug << oSeq[i] << " ";
	debug << endl;
	debug.close();
	#endif
	

	tlen=templates[tIndex]->length; 
	
	if(params.wwidth==INT_MAX) params.wwidth=maxint(tlen+1,nSeq+1);

	if(abs(nSeq-tlen)>params.wwidth)
		params.wwidth=abs(nSeq-tlen);
		
	if(params.dmetrics==DTW_DMETRICS) 
	{
		//If the selected distance metrics is derivative metrics, calculate the 
		//estimated derivatives of the two vectors
		thisobs=calcDerivatives(oSeq,nSeq);
		thistemp=calcDerivatives(templates[tIndex]->values,tlen);
	} //if
	else
	{
		//Otherwise, set observation and template pointers point to the original vectors
		thisobs=oSeq;
		thistemp=templates[tIndex]->values;
	} //else
	
	avgcost=FLT_MAX; 
	//The following two lines iterate through every possible start point of the path
	for(x=0;x<=params.seoffset;x++)
		for(y=maxint(0,x-params.wwidth);y<=minint(params.seoffset,x+params.wwidth);y++)
		{

			//Calculate a warping grid using the (x,y) as start point
			warpinggrid=createWarpingGrid(x,y,nSeq,tlen,params.wwidth,thisobs,thistemp,sqrDist);
						
			thiscost=FLT_MAX;
			//The following two lines iterate through every possible end point
			for(i=maxint(x,nSeq-params.seoffset-1);i<nSeq;i++)
				for(j=maxint3(tlen-params.seoffset-1,i-params.wwidth,y);j<minint(tlen,i+params.wwidth+1);j++)
					if(warpinggrid[i][j].costsofar/warpinggrid[i][j].lengthsofar<thiscost)
					{
						//If this end point is the best end point so far, save the coordinates
						//of end point
						thiscost=warpinggrid[i][j].costsofar/warpinggrid[i][j].lengthsofar;
						a=i;
						b=j;
					} //if
			if(thiscost<avgcost)
			{
				//If this start point is the best so far, save the average distance and the
				//coordinates of start and end points
				avgcost=thiscost;
				sx=x;
				sy=y;
				sa=a;
				sb=b;
			} //if

			for(i=0;i<nSeq;i++)
				delete[] warpinggrid[i];
			delete[] warpinggrid;
			warpinggrid=NULL;

			#ifdef DTWLIBDEBUG
			ofstream debug("componedebug.txt",ios_base::out|ios_base::app);
			debug << "x=" << x << ", y=" << y << ", a=" << a << ", b=" << b << ", thiscost=" << thiscost << endl;
			debug.close();
			#endif

		} //for

	//recreate the warping grid with best possible start point
	warpinggrid=createWarpingGrid(sx,sy,nSeq,tlen,params.wwidth,thisobs,thistemp,sqrDist);
	
	//The following code constructs the warping path into warpingpath STL vector
	temppoint.x=sa;
	temppoint.y=sb;
	while((temppoint.x!=sx)||(temppoint.y!=sy))
	{
		#ifdef DTWLIBDEBUG
		ofstream debug("componedebug.txt",ios_base::out|ios_base::app);
		debug << "WP.x=" << temppoint.x << ", WP.y=" << temppoint.y << endl;
		debug.close();
		#endif

		warpingpath.push_back(temppoint);
		i=temppoint.x;
		temppoint.x=warpinggrid[temppoint.x][temppoint.y].previ;
		temppoint.y=warpinggrid[i][temppoint.y].prevj;
	}//while
	warpingpath.push_back(temppoint);

	rplen=warpingpath.size();
	
	//Copy the contents of the STL vector into an array of size (rplen*2)
	retpath=new int*[rplen];
	for(i=0;i<rplen;i++)
	{
		retpath[i]=new int[2];
		retpath[i][0]=warpingpath[rplen-i-1].x;
		retpath[i][1]=warpingpath[rplen-i-1].y;
	} //for

	for(i=0;i<nSeq;i++)
		delete[] warpinggrid[i];
	delete[] warpinggrid;
	warpinggrid=NULL;

	if(params.dmetrics==DTW_DMETRICS) 
	{
		//Free the memory used by the two dynamic arrays created with calcDerivatives function
		delete[] thistemp;
		delete[] thisobs;
	} //if
	thistemp=NULL;
	thisobs=NULL;

	retresult=new DTWOneResult;
	retresult->dtwDistance=avgcost;
	retresult->pathLen=rplen;
	retresult->warpingPath=retpath;


	lastoneresult=retresult;
	return retresult;

  } //DTCompareAll


DTWLIB_API float* DTGetTemplate(int tIndex, int* nSeq)
//Returns the selected template
//
//Parameters:
//int tIndex - index number of the template
//int* nSeq - length of the returned template will be stored into this memory location
//
//Return value:
//float* - an array containing the requested template
{
	DTWTemplate *seltemplate; //A pointer to the requested template
	int i; //An iteration variable
	float * return_array = NULL; //If the template doesn't exist, a NULL pointer will be returned
    
	if((tIndex < templates.size()) && (templates[tIndex] != NULL))
	{
		//Set seltemplate point to the requested template structure
		seltemplate=templates[tIndex];
		return_array = new float[seltemplate->length];
		//Copy the contents of the requested template into return_array
		for (i=0; i < seltemplate->length; i++)
			return_array[i]=seltemplate->values[i];
		*nSeq=seltemplate->length;
	} //if
	else
	{
		*nSeq=0;
	} //else

	lasttemplate=return_array;
	return return_array;
} //DTGetTemplate


DTWLIB_API int DTFreeResources(void)
//Frees all the memory reserved by this DLL
{
	int i;

	delete[] lastallresults;
	lastallresults=NULL;

	delete[] lasttemplate;
	lasttemplate=NULL;

	if (lastoneresult!=NULL) {
		
		for(i=0;i<lastoneresult->pathLen;i++)
		{
			delete[] lastoneresult->warpingPath[i];
			lastoneresult->warpingPath[i]=NULL;
		}
		
		delete[] lastoneresult->warpingPath;
		lastoneresult->warpingPath=NULL;

		delete lastoneresult;
		lastoneresult=NULL;
	}

	return 0;
}

WarpingGridPoint** createWarpingGrid(int x, int y, int n, int m, int o, float* observation, 
									 float* ttemplate, float (*dist)(int,int,float*,float*))
//Calculates the warping grid
//
//Parameters:
//int x, int y: start point of the warping path
//int n: length of the observation array
//int m: length of the template array
//int o: warping window "width"
//float *observation: observation array
//float *ttemplate: template array
//float (*dist): a function pointer to the distance metrics function
//
//Return value:
//WarpingGridPoint** - the warping grid
{
	WarpingGridPoint** warpinggrid; //The warping grid
	int i,j; //Iteration variables. (i,j)=the point of warping grid that is being calculated
	int besti,bestj; //Temporary variables for storing the best one of possible previous points on warping path
    float bestavg; //Best cost of possible previous points
	float tempavg; //Current cost

	//Create new warping grid of size n*m
	warpinggrid = new WarpingGridPoint*[n];
	for(i=0;i<n;i++)
		warpinggrid[i]=new WarpingGridPoint[m];
	
	#ifdef DTWLIBDEBUG
	ofstream debug("wgdebug.txt",ios_base::out|ios_base::app);
	debug << "x=" << x << ", y=" << y << ", n=" << n << ", m=" << m << ", o=" << o <<endl;
	debug << "Grid created." << endl;
	debug.close();
	#endif
	
	//The total cost for the first point is the bare distance
	warpinggrid[x][y].costsofar=dist(x,y,observation,ttemplate);
	warpinggrid[x][y].lengthsofar=1;

	//Following two lines iterate through every possible point on warping path
	for(i=x;i<n;i++) { 
		for(j=maxint(y,i-o);j<minint(m,i+o+1);j++) {
			if ((i==x)&&(j==y))
				continue; //We have already dealt with the first point
			
			//Following three if blocks search the best one of possible previous
			//points on the warping path
			if ((i-1 >= x) && (abs((i-1)-j)<=o))
			{
				bestavg=( warpinggrid[i-1][j].costsofar + dist(i,j,observation,ttemplate) ) /
					(warpinggrid[i-1][j].lengthsofar+1);
				besti=i-1;
				bestj=j;
			} //if
			else	
				bestavg=FLT_MAX;

			if ((i-1 >= x) && (j-1 >= y))
			{
				tempavg=( warpinggrid[i-1][j-1].costsofar + dist(i,j,observation,ttemplate) ) /
					(warpinggrid[i-1][j-1].lengthsofar+1);
				if(tempavg<bestavg)
				{
					bestavg=tempavg;
					besti=i-1;
					bestj=j-1;
				} //if
			} //if

			if ((j-1 >= y) && (abs(i-(j-1))<=o))
			{
				tempavg=( warpinggrid[i][j-1].costsofar + dist(i,j,observation,ttemplate) ) /
					(warpinggrid[i][j-1].lengthsofar+1);
				if(tempavg<bestavg)
				{
					bestavg=tempavg;
					besti=i;
					bestj=j-1;
				} //if
			} //if
			
			//Calculate the total cost and length for this point and save information
			//about the previous point for reconstructing the warping path
			warpinggrid[i][j].costsofar=warpinggrid[besti][bestj].costsofar+dist(i,j,observation,ttemplate);
			warpinggrid[i][j].lengthsofar=warpinggrid[besti][bestj].lengthsofar+1;
			warpinggrid[i][j].previ=besti;
			warpinggrid[i][j].prevj=bestj;

#ifdef DTWLIBDEBUG
			ofstream debug("wgdebug.txt",ios_base::out|ios_base::app);
			debug << "i=" << i << ", j=" << j 
				<< ", costsofar=" <<warpinggrid[i][j].costsofar
				<< ", lengthsofar=" << warpinggrid[i][j].lengthsofar
				<< ", previ=" << besti << ", prevj= " << bestj << endl;
			debug.close();
#endif

		} //for
	} //for

	return warpinggrid;
} //CreateWarpingGrid

float sqrDist(int i, int j, float* observation, float* ttemplate) 
//Returns the euclidian distance between observation[i] and ttemplate[j]
{
	float a=observation[i]-ttemplate[j];
	return a*a;
} //sqrDist

float* calcDerivatives(float* sSeq, int slen)
//Returns an array containing estimated derivatives of time series sSeq
//
//Parameters:
//float *sSeq: the time series
//int slen: length of sSeq
//
//Return value:
//float*: an array of length slen containing the estimated derivatives
{
	int i; 
	float* deriv=new float[slen];

	for(i=1;i<slen-1;i++)
		deriv[i]=((sSeq[i]-sSeq[i-1])+(sSeq[i+1]-sSeq[i-1])/2)/2;
	
	//The derivatives can't be calculated for the first and last point so the
	//values for second and penultimate point are used instead
	deriv[0]=deriv[1];
	deriv[slen-1]=deriv[slen-2];

#ifdef DTWLIBDEBUG 
	ofstream debug("derivdebug.txt",ios_base::out|ios_base::app);
	for(i=0;i<slen;i++)
		debug << "i=" << i << ", Seq=" << sSeq[i] << ", deriv=" << deriv[i] << endl; 
	debug.close();
#endif	

	return deriv;
} //calcDerivatives

bool allResultsElementCompare(DTWAllResultsElement res1, DTWAllResultsElement res2)
//A comparison function to sort the results in DTCompareAll function
{
	return (res1.dtwDistance < res2.dtwDistance);
}




